#!/usr/bin/python

"""
prune_objectmap.py - script to prune object map files of squish test suites
    Initial version by Frerich Raabe Feb 2018
    Updated to support all script languages and SOM by Alan Ezust April 2020

    Requires Python 3 on Windows.

usage:
    prune_objectmap.py [suitedir] [objectMap-filename]

    If only one argument is specified, that is the test suite, or a directory of suites/scripts.
      Defaults to the current working directory
    If a second argument is specified, that is the object map filename.
      The object map file will be found automatically if the first argument is a valid test suite.

"""

import re
import os
import sys


class ObjectMap:
    """ Language-specific properties for reading/writing object map files """
    extension = ".map"
    delim = "\t"
    filename = "objects.map"

    def isOmLine(self, line):
        return self.delim in line

    def extract(self, line):
        [k, v] = line.split(self.delim, 1)
        return [k.strip(), v.strip()]


class Py_SOM(ObjectMap):
    extension = ".py"
    delim = "="
    filename = "names.py"

class JS_SOM(ObjectMap):
    extension = ".js"
    filename = "names.js"

    def isOmLine(self, line):
        return line.startswith("export var ")

    def extract(self, line):
        [k, v] = line[11:].split("=", 1)
        return [k.strip(), v.strip()]

class Rb_SOM(ObjectMap):
    extension = ".rb"
    delim = "="
    filename = "names.rb"

class Tcl_SOM(ObjectMap):
    extension = ".tcl"
    filename = "names.tcl"

    def isOmLine(self, line):
        return line.startswith("set ")
    def extract(self, line):
        [k, v] = re.split("\s+", line[4:], 1)
        return [k.strip(), v.strip()]

class Perl_SOM(ObjectMap):
    extension = ".pl"
    filename = "names.pl"

    def isOmLine(self, line):
        return line.startswith("our $")

    def extract(self, line):
        [k, v] = line[5:].split("=", 1)
        return [k.strip(), v.strip()]


oms = [ Py_SOM(), JS_SOM(), Rb_SOM(), Tcl_SOM(), Perl_SOM(), ObjectMap()]

def om_factory(filename):
    for om in oms:
        if filename.endswith(om.extension):
            return om
    print ("FATAL: No OM found for {}".format(filename))
    return False


def save_objectmap(fn, objectmap):
    """ writes out the object map file, by filtering out the
        entries that do not appear in the objectmap. """

    backupFile = "{}.bak".format(fn)
    os.rename(fn, backupFile)
    om = om_factory(fn)
    outF = open(fn, 'w')
    inF = open(backupFile, 'r')
    for line in inF:
        if om.isOmLine(line):
            [key, value] = om.extract(line)
            if key in objectmap:
                outF.write(line)
            else:
                print ("Pruned: {}".format(key))
        else:
            outF.write(line)
    outF.close()


def load_objectmap(fn):
    """Loads the object map of a Squish test suite. The object map
    file is referenced by the given file name. A dictionary is returned,
    containing the mapping of symbolic to real names.
    """
    result = {}
    om = om_factory(fn)

    with open(fn, 'r') as f:
        for line in f:
            if om.isOmLine(line):
                sn, rn = om.extract(line)
                result[sn] = rn
    return result


def files_beneath(d):
    """Yields a generator providing all files beneath a given directory. The
    traversal includes sub directories.
    """
    for path, _, files in os.walk(d):
        for filename in files:
            yield os.path.join(path, filename)


def file_contains(fn, s):
    """Returns True if the file 'fn' contains the string 's',
    False otherwise."""
    with open(fn, 'r') as f:
        return s in f.read()


def symbolic_name_used(script_files, objectmap, symbolic_name):
    """Returns True if the given symbolic name is referenced in either any
    of the given script files or in any of the real names of the given object
    map.
    """
    return any(file_contains(fn, symbolic_name) for fn in script_files) \
        or any(symbolic_name in rn for _, rn in objectmap.items())


def prune_objectmap(script_files, objectmap):
    """Removes all entries in the given objct map which are not referenced by any
    of the given script files; container objects which are rendered unused after
    removing an object name will be removed, too. Returns the new pruned object
    map.
    """

    # TODO: Inefficient way to detect unused containers; a better version would
    # construct a directed graph and then perform a topological sort to find
    # leaves (and their parents).
    while True:
        unused_names = set(sn for sn, _ in objectmap.items() if not symbolic_name_used(script_files, objectmap, sn))
        if not unused_names:
            break
        objectmap = {sn: rn for sn, rn in objectmap.items() if not sn in unused_names}
    return objectmap

def find_SOM(path):
    for f in os.listdir(os.path.join(path, "shared", "scripts")):
        for om in oms:
            if f == om.filename:
                return os.path.join(path, "shared", "scripts", f)
    return False


if __name__ == '__main__':

    testsuite_dir = "."
    if len(sys.argv) > 1:
        testsuite_dir = sys.argv[1]

    objectmap_filename = os.path.join(testsuite_dir, "objects.map")
    if len(sys.argv) > 2:
        objectmap_filename = sys.argv[2]
        print ("Using objectmap file: {}".format(objectmap_filename))

    if not os.path.exists(objectmap_filename):
        objectmap_filename = find_SOM(testsuite_dir)
        if not objectmap_filename:
            print ("No object map file found. ")
            sys.exit(-1)

    # Scan all test suite files including .xml (verification points)
    script_files = [fn for fn in files_beneath(testsuite_dir) if not os.path.samefile(fn, objectmap_filename)]

    objectmap = load_objectmap(objectmap_filename)
    objectmap_new = prune_objectmap(script_files, objectmap)
    if len(objectmap) == len(objectmap_new):
        print ("No unused entries found. Nothing to do.")
    elif len(objectmap_new) == 0:
        print ("??? {}: No object names are used".format(objectmap_filename))
    else:
        save_objectmap(objectmap_filename, objectmap_new)
        print("Pruned {} from {} down to {}".format(objectmap_filename, len(objectmap), len(objectmap_new)))


